/* mpn_mul_1 -- Multiply a limb vector with a single limb and store the
   product in a second limb vector.

Copyright 1991, 1992, 1993, 1994, 1995, 1996, 1997, 2000, 2001, 2002, 2005 Free Software
Foundation, Inc.

This file is part of the GNU MP Library.

The GNU MP Library is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or (at your
option) any later version.

The GNU MP Library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

You should have received a copy of the GNU Lesser General Public License
along with the GNU MP Library; see the file COPYING.LIB.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
MA 02111-1307, USA. */

#include "gmp.h"
#include "gmp-impl.h"
#include "longlong.h"

// INPUT PARAMETERS
// res_ptr	r3
// s1_ptr	r4
// size     r5
// s2_limb	r6
// OUTPUT
// cy_limb      r3

// converted to 64-bit from 32-bit mul_1.asm routine in mpn/powerpc32
// with some rearrangement

mp_limb_t
mpn_mul_1 (mp_ptr rp, mp_srcptr up, mp_size_t n, mp_limb_t vl)
{
 asm {
	ld      r8,0(r4)
	mtctr	r5
	mulld	r11,r8,r6
	bdz     @L_end3     // branch if length = 1
	addic	r0,r0,0		// clear carry bit
	ldu     r9,8(r4)
	mulhdu	r0,r8,r6
	addi	r3,r3,-8	// adjust res_ptr, it's offset before it's used
	li      r12,0		// clear upper product reg
	bdz     @L_end1     // branch if length = 2
// Software pipelined main loop
@Loop:
	ld      r8,8(r4)
	mulld	r10,r9,r6
	adde	r5,r11,r12
	mulhdu	r12,r9,r6
	std     r5,8(r3)
	bdz     @L_end2
	ldu     r9,16(r4)
	mulld	r11,r8,r6
	adde	r7,r10,r0
	mulhdu	r0,r8,r6
	stdu	r7,16(r3)
	bdnz	@Loop
// Finish software pipeline
@L_end1:
	mulld	r10,r9,r6
	adde	r5,r11,r12
	mulhdu	r12,r9,r6
	std     r5,8(r3)
	adde	r7,r10,r0
	stdu	r7,16(r3)
	addze	r3,r12
	blr
@L_end2:
	mulld	r11,r8,r6
	adde	r7,r10,r0
	mulhdu	r0,r8,r6
	stdu	r7,16(r3)
	adde	r5,r11,r12
	std     r5,8(r3)
	addze	r3,r0
	blr
@L_end3:
	std     r11,0(r3)
	mulhdu	r3,r8,r6
	blr
 }
}

// INPUT PARAMETERS
// res_ptr	r3
// s1_ptr	r4
// size     r5
// s2_limb	r6
// cy_limb  r7
// OUTPUT
// cy       r3

mp_limb_t
mpn_mul_1c (mp_ptr rp, mp_srcptr up, mp_size_t n, mp_limb_t vl, mp_limb_t cy)
{
 asm {
	ld      r8,0(r4)
	mtctr	r5
	mulld	r11,r8,r6
	bdz     @L_end3     // branch if length = 1
	addic	r0,r0,0		// clear carry bit
	ldu     r9,8(r4)
	mulhdu	r0,r8,r6
	addi	r3,r3,-8	// adjust res_ptr, it's offset before it's used
	mr      r12,r7		// limb carried in
	bdz     @L_end1    // branch if length = 2
// Software pipelined main loop
@Loop:
	ld      r8,8(r4)
	mulld	r10,r9,r6
	adde	r5,r11,r12
	mulhdu	r12,r9,r6
	std     r5,8(r3)
	bdz     @L_end2
	ldu     r9,16(r4)
	mulld	r11,r8,r6
	adde	r7,r10,r0
	mulhdu	r0,r8,r6
	stdu	r7,16(r3)
	bdnz	@Loop
// Finish software pipeline
@L_end1:
	mulld	r10,r9,r6
	adde	r5,r11,r12
	mulhdu	r12,r9,r6
	std     r5,8(r3)
	adde	r7,r10,r0
	stdu	r7,16(r3)
	addze	r3,r12
	blr
@L_end2:
	mulld	r11,r8,r6
	adde	r7,r10,r0
	mulhdu	r0,r8,r6
	stdu	r7,16(r3)
	adde	r5,r11,r12
	std     r5,8(r3)
	addze	r3,r0
	blr
@L_end3:
	mulhdu	r5,r8,r6
    addc    r11,r11,r7
	std     r11,0(r3)
    addze   r3,r5
	blr
 }
}





