/* mpn_submul_1 -- multiply the N long limb vector pointed to by UP by VL,
   subtract the N least significant limbs of the product from the limb
   vector pointed to by RP.  Return the most significant limb of the
   product, adjusted for carry-out from the subtraction.

Copyright 1992, 1993, 1994, 1995, 1996, 1997, 1998, 2000, 2002, 2005 Free Software
Foundation, Inc.

This file is part of the GNU MP Library.

The GNU MP Library is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or (at your
option) any later version.

The GNU MP Library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

You should have received a copy of the GNU Lesser General Public License
along with the GNU MP Library; see the file COPYING.LIB.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
MA 02111-1307, USA. */

#include "gmp.h"
#include "gmp-impl.h"
#include "longlong.h"

// converted to 64-bit from 32-bit submul_1.asm routine in mpn/powerpc32
// with some rearrangement

mp_limb_t
mpn_submul_1 (mp_ptr rp, mp_srcptr up, mp_size_t n, mp_limb_t vl)
{
 asm {
	cmpdi	r5,9        // more than 9 limbs?
	bgt     @Lbig       // branch if more than 9 limbs

	ld      r0,0(r4)
	mtctr	r5
	ld      r9,0(r3)
	mulld	r7,r0,r6
	addi	r3,r3,-8
	mulhdu	r10,r0,r6
	subfc	r8,r7,r9
	addc	r7,r7,r8	// invert cy (r7 is junk)
	bdz     @Lend
@Loop:
	ldu     r0,8(r4)
	stdu	r8,8(r3)
	mulld	r8,r0,r6
	ld      r9,8(r3)
	mulhdu	r11,r0,r6
	adde	r7,r8,r10
	addze	r10,r11
	subfc	r8,r7,r9
	addc	r7,r7,r8	// invert cy (r7 is junk)
	bdnz	@Loop
@Lend:
	std     r8,8(r3)
	addze	r3,r10
	blr

@Lbig:
	std     r30,-40(r1)
	std     r31,-32(r1)
	addi	r5,r5,-1
	srdi	r0,r5,2
	mtctr	r0

	ld      r7,0(r4)
	mulld	r8,r7,r6
	mulhdu	r0,r7,r6
	ld      r7,0(r3)
	subfc	r7,r8,r7
	addc	r8,r8,r7
	std     r7,0(r3)

@LoopU:
	ld      r7,8(r4)
	ld      r12,16(r4)
	ld      r30,24(r4)
	ldu     r31,32(r4)
	mulld	r8,r7,r6
	mulld	r9,r12,r6
	mulld	r10,r30,r6
	mulld	r11,r31,r6
	adde	r8,r8,r0	// add cy_limb
	mulhdu	r0,r7,r6
	ld      r7,8(r3)
	adde	r9,r9,r0
	mulhdu	r0,r12,r6
	ld      r12,16(r3)
	adde	r10,r10,r0
	mulhdu	r0,r30,r6
	ld      r30,24(r3)
	adde	r11,r11,r0
	mulhdu	r0,r31,r6
	ld      r31,32(r3)
	addze	r0,r0		// new cy_limb
	subfc	r7,r8,r7
	std     r7,8(r3)
	subfe	r12,r9,r12
	std     r12,16(r3)
	subfe	r30,r10,r30
	std     r30,24(r3)
	subfe	r31,r11,r31
	stdu	r31,32(r3)
	subfe	r11,r11,r11     // invert ...
	addic	r11,r11,1       // ... carry
	bdnz	@LoopU

	andi.	r31,r5,3
	mtctr	r31
	beq     cr0,@Lendx

@LoopE:
	ldu     r7,8(r4)
	mulld	r8,r7,r6
	adde	r8,r8,r0	// add cy_limb
	mulhdu	r0,r7,r6
	ld      r7,8(r3)
	addze	r0,r0		// new cy_limb
	subfc	r7,r8,r7
	addc	r8,r8,r7
	stdu	r7,8(r3)
	bdnz	@LoopE
@Lendx:
	addze	r3,r0
	ld     r30,-40(r1)
	ld     r31,-32(r1)
	blr
 }
}

mp_limb_t
mpn_submul_1c (mp_ptr rp, mp_srcptr up, mp_size_t n, mp_limb_t vl, mp_limb_t cy)
{
 asm {
	cmpdi	r5,9        // more than 9 limbs?
	bgt     @Lbig       // branch if more than 9 limbs

	ld      r0,0(r4)
	mtctr	r5
	ld      r9,0(r3)
	mulld	r8,r0,r6
	addi	r3,r3,-8
	mulhdu	r10,r0,r6
    addc    r7,r8,r7
    addze   r10,r10
	subfc	r8,r7,r9
	addc	r7,r7,r8	// invert cy (r7 is junk)
	bdz     @Lend
@Loop:
	ldu     r0,8(r4)
	stdu	r8,8(r3)
	mulld	r8,r0,r6
	ld      r9,8(r3)
	mulhdu	r11,r0,r6
	adde	r7,r8,r10
	addze	r10,r11
	subfc	r8,r7,r9
	addc	r7,r7,r8	// invert cy (r7 is junk)
	bdnz	@Loop
@Lend:
	std     r8,8(r3)
	addze	r3,r10
	blr

@Lbig:
	std     r30,-40(r1)
	std     r31,-32(r1)
	addi	r5,r5,-1
	srdi	r0,r5,2
	mtctr	r0

	ld      r9,0(r4)
	mulld	r8,r9,r6
	mulhdu	r0,r9,r6
    addc    r8,r8,r7
    addze   r0,r0
	ld      r7,0(r3)
	subfc	r7,r8,r7
	addc	r8,r8,r7
	std     r7,0(r3)

@LoopU:
	ld      r7,8(r4)
	ld      r12,16(r4)
	ld      r30,24(r4)
	ldu     r31,32(r4)
	mulld	r8,r7,r6
	mulld	r9,r12,r6
	mulld	r10,r30,r6
	mulld	r11,r31,r6
	adde	r8,r8,r0	// add cy_limb
	mulhdu	r0,r7,r6
	ld      r7,8(r3)
	adde	r9,r9,r0
	mulhdu	r0,r12,r6
	ld      r12,16(r3)
	adde	r10,r10,r0
	mulhdu	r0,r30,r6
	ld      r30,24(r3)
	adde	r11,r11,r0
	mulhdu	r0,r31,r6
	ld      r31,32(r3)
	addze	r0,r0		// new cy_limb
	subfc	r7,r8,r7
	std     r7,8(r3)
	subfe	r12,r9,r12
	std     r12,16(r3)
	subfe	r30,r10,r30
	std     r30,24(r3)
	subfe	r31,r11,r31
	stdu	r31,32(r3)
	subfe	r11,r11,r11     // invert ...
	addic	r11,r11,1       // ... carry
	bdnz	@LoopU

	andi.	r31,r5,3
	mtctr	r31
	beq     cr0,@Lendx

@LoopE:
	ldu     r7,8(r4)
	mulld	r8,r7,r6
	adde	r8,r8,r0	// add cy_limb
	mulhdu	r0,r7,r6
	ld      r7,8(r3)
	addze	r0,r0		// new cy_limb
	subfc	r7,r8,r7
	addc	r8,r8,r7
	stdu	r7,8(r3)
	bdnz	@LoopE
@Lendx:
	addze	r3,r0
	ld     r30,-40(r1)
	ld     r31,-32(r1)
	blr
 }
}
